/*!
  @file rx_mc.h
	
  @brief A֐\ʂ̃|S(MC@)
	
	http://local.wasp.uwa.edu.au/~pbourke/geometry/polygonise/
 
  @author Raghavendra Chandrashekara (basesd on source code
			provided by Paul Bourke and Cory Gene Bloyd)
  @date   2010-03
*/


#ifndef _RX_MC_MESH_H_
#define _RX_MC_MESH_H_


//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------
// CW
#include <cstdlib>

// OpenGL
#include <GL/glew.h>
#include <GL/glut.h>

// STL
#include <map>
#include <vector>
#include <string>

#include <iostream>

#include "rx_utility.h"
#include "rx_mesh.h"


//-----------------------------------------------------------------------------
// O
//-----------------------------------------------------------------------------
using namespace std;


//-----------------------------------------------------------------------------
// `
//-----------------------------------------------------------------------------
typedef unsigned int uint;

#ifndef RXREAL
	#define RXREAL float
#endif


struct RxVertexID
{
	uint newID;
	double x, y, z;
};

typedef std::map<uint, RxVertexID> ID2VertexID;

struct RxTriangle
{
	uint pointID[3];
};

typedef std::vector<RxTriangle> RxTriangleVector;

struct RxScalarField
{
	uint iNum[3];
	Vec3 fWidth;
	Vec3 fMin;
};



//-----------------------------------------------------------------------------
// rxMCMeshCPUNX
//-----------------------------------------------------------------------------
class rxMCMeshCPU
{
public:
	// RXgN^
	rxMCMeshCPU();

	// fXgN^
	~rxMCMeshCPU();
	
	//! A֐Op`bV𐶐
	bool CreateMeshF(RXREAL (*func)(double, double, double), Vec3 min_p, double h, int n[3], RXREAL threshold, 
					 vector<Vec3> &vrts, vector<Vec3> &nrms, vector<rxFace> &face);

	//! Tv{[Op`bV𐶐
	bool CreateMeshV(RXREAL *field, Vec3 min_p, double h, int n[3], RXREAL threshold, 
					 vector<Vec3> &vrts, vector<Vec3> &nrms, vector<rxFace> &face);

	//! Tv{[瓙lʃbV
	void GenerateSurface(const RxScalarField sf, RXREAL *field, RXREAL threshold, 
						 vector<Vec3> &vrts, vector<Vec3> &nrms, vector<int> &tris);

	//! ֐Tv{[쐬ēlʃbV
	void GenerateSurfaceV(const RxScalarField sf, RXREAL (*func)(double, double, double), RXREAL threshold, 
						  vector<Vec3> &vrts, vector<Vec3> &nrms, vector<int> &tris);

	//! ֐瓙lʃbV
	void GenerateSurfaceF(const RxScalarField sf, RXREAL (*func)(double, double, double), RXREAL threshold, 
						  vector<Vec3> &vrts, vector<Vec3> &nrms, vector<int> &tris);

	//! lʍ쐬trueԂ
	bool IsSurfaceValid() const { return m_bValidSurface; }

	//! 쐬lʃbV̔j
	void DeleteSurface();

	//! bVɗpObh̑傫(bV쐬ĂȂꍇ͕Ԓl-1)
	int GetVolumeLengths(double& fVolLengthX, double& fVolLengthY, double& fVolLengthZ);

	// 쐬bV̏
	uint GetNumVertices(void) const { return m_nVertices; }
	uint GetNumTriangles(void) const { return m_nTriangles; }
	uint GetNumNormals(void) const { return m_nNormals; }

protected:
	// MARK:oϐ
	uint m_nVertices;	//!< lʃbV̒_
	uint m_nNormals;	//!< lʃbV̒_@(쐬Ă @=_)
	uint m_nTriangles;	//!< lʃbV̎Op`|S

	ID2VertexID m_i2pt3idVertices;			//!< lʂ`钸_̃Xg
	RxTriangleVector m_trivecTriangles;		//!< Op`|S`钸_̃Xg

	RxScalarField m_Grid;					//!< Obh

	// A֐l(XJ[l)擾pϐ(ǂ炩̂ݗp)
	const RXREAL* m_ptScalarField;				//!< XJ[lێTv{[
	RXREAL (*m_fpScalarFunc)(double, double, double);	//!< XJ[lԂ֐|C^

	RXREAL m_tIsoLevel;							//!< 臒l

	bool m_bValidSurface;					//!< bV̉


	// bV\zp̃e[u
	static const uint m_edgeTable[256];
	static const int m_triTable[256][16];



	// MARK:protectedo֐

	//! GbWID
	uint GetEdgeID(uint nX, uint nY, uint nZ, uint nEdgeNo);

	//! _ID
	uint GetVertexID(uint nX, uint nY, uint nZ);

	// GbW̓l_vZ
	RxVertexID CalculateIntersection(uint nX, uint nY, uint nZ, uint nEdgeNo);
	RxVertexID CalculateIntersectionF(uint nX, uint nY, uint nZ, uint nEdgeNo);

	//! ObhGbW[̉A֐l^Ԃœl_vZ
	RxVertexID Interpolate(double fX1, double fY1, double fZ1, double fX2, double fY2, double fZ2, RXREAL tVal1, RXREAL tVal2);

	//! _CbV􉽏o͌`Ŋi[
	void RenameVerticesAndTriangles(vector<Vec3> &vrts, uint &nvrts, vector<int> &tris, uint &ntris);

	//! _@vZ
	void CalculateNormals(const vector<Vec3> &vrts, uint nvrts, const vector<int> &tris, uint ntris, 
						  vector<Vec3> &nrms, uint &nnrms);

};


//-----------------------------------------------------------------------------
// rxMCMeshGPUNX
//-----------------------------------------------------------------------------
class rxMCMeshGPU
{
protected:
	// MC@p
	uint m_u3GridSize[3];			//!< Obh(nx,ny,nz)
	uint m_u3GridSizeMask[3];		//!< Obh/CfbNXϊ̃}XN
	uint m_u3GridSizeShift[3];		//!< Obh/CfbNXϊ̃Vtg

	float m_f3VoxelMin[3];			//!< Obhŏʒu
	float m_f3VoxelMax[3];			//!< Obhőʒu
	float m_f3VoxelH[3];			//!< Obh
	uint m_uNumVoxels;				//!< Obh
	uint m_uMaxVerts;				//!< ő咸_
	uint m_uNumActiveVoxels;		//!< bV݂{NZ
	uint m_uNumVrts;				//!< _
	uint m_uNumTris;				//!< bV

	// foCX
	float *g_dfVolume;				//!< A֐f[^i[Obh
	float *g_dfNoise;				//!< mCYli[Obh(`掞̐F肷̂Ɏgp)
	uint *m_duVoxelVerts;			//!< ObhɊ܂܂郁bV_
	uint *m_duVoxelVertsScan;		//!< ObhɊ܂܂郁bV_(Scan)
	uint *m_duCompactedVoxelArray;	//!< bV܂ރObh

	uint *m_duVoxelOccupied;		//!< |Sɑ݂{NZ̃Xg
	uint *m_duVoxelOccupiedScan;	//!< |Sɑ݂{NZ̃Xg(prefix scan)

	// 􉽏𐶐ƂɕKvȕϐ
	uint m_u3EdgeSize[3][3];		//!< GbW(nx,ny,nz)
	uint m_uNumEdges[4];			//!< GbW

	// 􉽏𐶐ƂɕKvȕϐ
	uint *m_duVoxelCubeIdx;			//!< Obh8_̉A֐l臒lȏォǂerbgɊi[ϐ
	uint *m_duEdgeOccupied;			//!< GbWɃbV_܂ނǂ(xCy, z̏)
	uint *m_duEdgeOccupiedScan;		//!< GbWɃbV_܂ނǂ(Scan)
	float *m_dfEdgeVrts;			//!< GbWƂɎZo_
	float *m_dfCompactedEdgeVrts;	//!< Ԃ߂_
	uint *m_duIndexArray;			//!< |S̊􉽏
	float *m_dfVertexNormals;		//!< _@
	uint *m_duVoxelTriNum;			//!< ObhƂ̎Op`bV
	uint *m_duVoxelTriNumScan;		//!< ObhƂ̎Op`bV(Scan)

	// 􉽏KvƂȂƂ̂ݗp
	float *m_df4Vrts;				//!< |S_W
	float *m_df4Nrms;				//!< |S_@


	// zXg
	float *m_f4VertPos;			//!< _W
	uint *m_u3TriIdx;				//!< bVCfbNX
	uint *m_uScan;					//!< fobOp

	float *m_f4VertNrm;			//!< _@


	int m_iVertexStore;
	bool m_bSet;


	// A֐l(XJ[l)擾pϐ(ǂ炩̂ݗp)
	const float* m_ptScalarField;				//!< XJ[lێTv{[
	float (*m_fpScalarFunc)(double, double, double);	//!< XJ[lԂ֐|C^

	float m_tIsoLevel;							//!< 臒l

	bool m_bValidSurface;						//!< bV̉

public:
	// RXgN^
	rxMCMeshGPU();

	// fXgN^
	~rxMCMeshGPU();
	
	//! A֐Op`bV𐶐
	bool CreateMeshF(RXREAL (*func)(double, double, double), Vec3 min_p, double h, int n[3], float threshold, 
					 vector<Vec3> &vrts, vector<Vec3> &nrms, vector<rxFace> &face);

	//! Tv{[Op`bV𐶐
	bool CreateMeshV(RXREAL *volume, Vec3 minp, double h, int n[3], float threshold, 
					 vector<Vec3> &vrts, vector<Vec3> &nrms, vector<rxFace> &face);

	//! Tv{[Op`bV𐶐
	bool CreateMeshV(Vec3 minp, double h, int n[3], float threshold, uint &nvrts, uint &ntris);
	
	//! z̊m
	bool Set(Vec3 minp, Vec3 h, int n[3], unsigned int vm = 5);

	//! z̍폜
	void Clean(void);

	//! FBOɃf[^ݒ
	bool SetDataToFBO(GLuint uVrtVBO, GLuint uNrmVBO, GLuint uTriVBO);

	//! zXgzɃf[^ݒ
	bool SetDataToArray(vector<Vec3> &vrts, vector<Vec3> &nrms, vector<rxFace> &face);

	//! Tv{[Zbg
	void   SetSampleVolumeFromHost(float *hVolume);
	float* GetSampleVolumeDevice(void);

	void   SetSampleNoiseFromHost(float *hVolume);
	float* GetSampleNoiseDevice(void);

	// ő咸_
	uint GetMaxVrts(void){ return m_uMaxVerts; }

	// ő咸_vZpW
	void SetVertexStore(int vs){ m_iVertexStore = vs; }

	//! _f[^(foCX)
	float* GetVrtDev(void);

	//! bVf[^(foCX)
	uint* GetIdxDev(void);

	//! @f[^(foCX)
	float* GetNrmDev(void);

	//! _f[^(zXg)
	float GetVertex(int idx, int dim)
	{
		if(dim == 0){
			return m_f4VertPos[4*idx+0];
		}
		else if(dim == 1){
			return m_f4VertPos[4*idx+1];
		}
		else{
			return m_f4VertPos[4*idx+2];
		}
	}
	void GetVertex2(int idx, float *x, float *y, float *z)
	{
		*x = m_f4VertPos[4*idx+0];
		*y = m_f4VertPos[4*idx+1];
		*z = m_f4VertPos[4*idx+2];
	}

	//! bVf[^(zXg)
	void GetTriIdx(int idx, unsigned int *vidx0, unsigned int *vidx1, unsigned int *vidx2)
	{
		*vidx0 = m_u3TriIdx[4*idx+0];
		*vidx1 = m_u3TriIdx[4*idx+1];
		*vidx2 = m_u3TriIdx[4*idx+2];
	}
};




#endif // _RX_MC_MESH_H_

