/*! 
  @file rx_cu_common.cuh
	
  @brief CUDAʃwb_
 
  @author Makoto Fujisawa
  @date 2009-08, 2011-06
*/
// FILE --rx_cu_common.cuh--

#ifndef _RX_CU_COMMON_CUH_
#define _RX_CU_COMMON_CUH_


//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------
#include "vector_types.h"
#include "vector_functions.h"


//-----------------------------------------------------------------------------
// `
//-----------------------------------------------------------------------------


typedef unsigned int uint;
typedef unsigned char uchar;

#define FLOAT float
#define FLOAT3 float3

#define MAKE_FLOAT3 make_float3

#define RX_CUMC_USE_GEOMETRY

#define RX_USE_ATOMIC_FUNC // vCompute capability 1.1ȏ(-arch sm_11)


// eNX`̎gptO
#ifndef __DEVICE_EMULATION__
#define USE_TEX 0
#endif

#if USE_TEX
#define FETCH(t, i) tex1Dfetch(t##Tex, i)
#else
#define FETCH(t, i) t[i]
#endif

#if USE_TEX
#define FETCHC(t, i) tex1Dfetch(t##Tex, i)
#else
#define FETCHC(t, i) cell.t[i]
#endif


// 1ubÑXbh(/)
#define BLOCK_SIZE 16

// 1ubN̍őXbh
#define THREAD_NUM 512

// Tv{[pꍇ1, A֐pꍇ0
#define SAMPLE_VOLUME 1

// Shared Memory̎gptO
#define USE_SHARED 1

// Shared Memorypꍇ̃Xbh̐
#define NTHREADS 32

#define SKIP_EMPTY_VOXELS 1



// s
struct matrix3x3
{
	float3 e[3];
};

struct matrix4x4
{
	float4 e[4];
};


#define MAX_POLY_NUM 10
#define MAX_BOX_NUM 10
#define MAX_SPHERE_NUM 10

#define DEBUG_N 17


//-----------------------------------------------------------------------------
//! SPHV~[Vp[^
//-----------------------------------------------------------------------------
struct rxSimParams
{
	FLOAT3 Gravity;
	FLOAT ParticleRadius;

	float3 Boundary[2];

	uint3 GridSize;
	uint NumCells;
	FLOAT3 WorldOrigin;
	FLOAT3 WorldMax;
	FLOAT3 CellWidth;

	uint NumBodies;
	uint MaxParticlesPerCell;

	uint3 GridSizeB;
	uint NumCellsB;
	FLOAT3 WorldOriginB;
	FLOAT3 WorldMaxB;
	FLOAT3 CellWidthB;
	uint NumBodiesB;


	FLOAT EffectiveRadius;
	FLOAT Mass;			// p[eBN[kg]
	FLOAT VorticityConfinement;

	FLOAT Buoyancy;

	FLOAT Density;		// x[kg/m^3]
	FLOAT Pressure;     // [Pa = N/m^2 = kg/m.s^2]

	FLOAT Tension;		// [N/m = kg/s^2]
	FLOAT Viscosity;	// [Pa.s = N.s/m^2 = kg/m.s]
	FLOAT GasStiffness;	// [J = N.m = kg.m^2/s^2]  // used for DC96 symmetric pressure force

	FLOAT Volume;
	FLOAT KernelParticles;
	FLOAT Restitution;

	FLOAT Threshold;

	FLOAT InitDensity;

	FLOAT Dt;

	FLOAT Wpoly6;		//!< Pory6J[l̒萔W
	FLOAT GWpoly6;		//!< Pory6J[ľz̒萔W
	FLOAT LWpoly6;		//!< Pory6J[l̃vVA̒萔W
	FLOAT Wspiky;		//!< SpikyJ[l̒萔W
	FLOAT GWspiky;		//!< SpikyJ[ľz̒萔W
	FLOAT LWspiky;		//!< SpikyJ[l̃vVA̒萔W
	FLOAT Wvisc;		//!< ViscosityJ[l̒萔W
	FLOAT GWvisc;		//!< ViscosityJ[ľz̒萔W
	FLOAT LWvisc;		//!< ViscosityJ[l̃vVA̒萔W

	FLOAT Wpoly6r;		//!< Pory6J[l̒萔W(ől1ɂȂ悤ɐKW)

	int AP;
	FLOAT AP_K;
	FLOAT AP_N;
	FLOAT AP_Q;
	FLOAT AP_WQ;

	uint   PolyNum;
	FLOAT3 PolyVel[MAX_POLY_NUM];

	uint   BoxNum;
#if MAX_BOX_NUM
	FLOAT3 BoxCen[MAX_BOX_NUM];
	FLOAT3 BoxExt[MAX_BOX_NUM];
	//FLOAT3 BoxVel[MAX_BOX_NUM];
	matrix3x3 BoxRot[MAX_BOX_NUM];
	matrix3x3 BoxInvRot[MAX_BOX_NUM];
	uint   BoxFlg[MAX_BOX_NUM];
#endif

	uint SphereNum;
#if MAX_SPHERE_NUM
	FLOAT3 SphereCen[MAX_SPHERE_NUM];
	FLOAT  SphereRad[MAX_SPHERE_NUM];
	//FLOAT3 SphereVel[MAX_SPHERE_NUM];
	uint   SphereFlg[MAX_SPHERE_NUM];
#endif
};



struct rxParticleCell
{
	float4* dSortedPos;			//!< \[gς݃p[eBNW
	float4* dSortedVel;			//!< \[gς݃p[eBNx

	uint* dSortedIndex;			//!< \[gς݃p[eBNCfbNX
	uint* dGridParticleHash;	//!< ep[eBÑObhnbVl(\[gpL[)
	uint* dCellStart;			//!< \[gXg̊eZ̃X^[gCfbNX
	uint* dCellEnd;				//!< \[gXg̊eZ̃GhCfbNX
	uint  uNumParticles;		//!< p[eBN
	uint  uNumCells;			//!< Z
	uint  uNumArdGrid;			//!< ߖTTQƃObh͈

	uint* dSortedPolyIdx;		//!< \[gς݃|SCfbNX
	uint* dGridPolyHash;		//!< |S̃ObhnbVl(\[gpL[)
	uint* dPolyCellStart;		//!< \[gXg̊eZ̃X^[gCfbNX
	uint* dPolyCellEnd;			//!< \[gXg̊eZ̃GhCfbNX
	uint  uNumPolyHash;
};





#endif // _RX_CU_COMMON_CUH_
