/*! 
  @file rx_cu_funcs.cu
	
  @brief CUDA֐ - ֌WȂ

  @author Makoto Fujisawa
  @date 2009-08, 2011-06
*/
// FILE --rx_cu_funcs.cu--


//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------
#include <cstdio>
#include <GL/glew.h>
#include <GL/glut.h>

#include <thrust/device_vector.h>
#include <thrust/scan.h>
#include <thrust/sort.h>

#include <GL/freeglut.h>

#include "rx_cu_common.cu"



//-----------------------------------------------------------------------------
// CUDA֐
//-----------------------------------------------------------------------------
extern "C"
{
/*!
 * CUDAfoCX̐ݒ
 *  - R}hCɊÂCUDAfoCXݒ(()-device 0)
 * @param[in] argc R}hC̐
 * @param[in] argv R}hCXg(argv[0]͎st@C)
 */
void CuInit(int argc, char **argv)
{   
	if(checkCmdLineFlag(argc, (const char**)argv, "device")){
		int id = getCmdLineArgumentInt(argc, (const char**)argv, "device=");
		if(id < 0){
			id = gpuGetMaxGflopsDeviceId();
			cudaSetDevice(id);
		}
		else{
			cudaSetDevice(id);
		}
	}
	else{
		cudaSetDevice( gpuGetMaxGflopsDeviceId() );
	}
}

/*!
 * CUDAfoCX̐ݒ - id𒼐ڎw
 * @param[in] id foCXID
 */
void CuSetDevice(int id)
{ 
	int device_count = 0;
	cudaGetDeviceCount(&device_count);
	if(id < 0 || id >= device_count){
		id = gpuGetMaxGflopsDeviceId();
		cudaSetDevice(0);
	}
	else{
		cudaSetDevice(id);
	}
}

/*!
 * foCX̊m
 * @param[out] dPtr foCXւ̃|C^
 * @param[in] size mۃTCY(̃TCY)
 */
void CuAllocateArray(void **dPtr, size_t size)
{
	RX_CUCHECK(cudaMalloc(dPtr, size));
}

/*!
 * foCX̉
 * @param[in] devPtr foCXւ̃|C^
 */
void CuFreeArray(void *dPtr)
{
	RX_CUCHECK(cudaFree(dPtr));
}

/*!
 * foCẌ̏
 * @param[in] dPtr foCXւ̃|C^
 * @param[in] val l
 * @param[in] size ̈̃TCY(̃TCY)
 */
void CuSetArrayValue(void *dPtr, int val, size_t size)
{
	RX_CUCHECK(cudaMemset(dPtr, val, size));
}

/*!
 * foCXԃRs[
 * @param[in] dDst Rs[
 * @param[in] dSrc Rs[
 * @param[in] size Rs[TCY(̃TCY)
 */
void CuCopyArrayD2D(void *dDst, void *dSrc, int size)
{
	RX_CUCHECK(cudaMemcpy(dDst, dSrc, size, cudaMemcpyDeviceToDevice));
}


/*!
 * VBO}bsO
 * @param[in] vbo VBO,PBO
 */
void *CuMapGLBufferObject(cudaGraphicsResource **resource)
{
	void *ptr;
	RX_CUCHECK(cudaGraphicsMapResources(1, resource, 0));
	size_t num_bytes;
	RX_CUCHECK(cudaGraphicsResourceGetMappedPointer((void**)&ptr, &num_bytes, *resource));
	return ptr;
}

/*!
 * VBOA}bv
 * @param[in] vbo VBO,PBO
 */
void CuUnmapGLBufferObject(cudaGraphicsResource *resource)
{
	RX_CUCHECK(cudaGraphicsUnmapResources(1, &resource, 0));
}

/*!
 * PBO,VBOobt@CUDAɓo^
 * @param[in] vbo VBO,PBO
 */
void CuRegisterGLBufferObject(uint vbo, cudaGraphicsResource **resource)
{
	RX_CUCHECK(cudaGraphicsGLRegisterBuffer(resource, vbo, cudaGraphicsMapFlagsNone));
}

/*!
 * PBO,VBOobt@CUDA폜
 * @param[in] vbo VBO,PBO
 */
void CuUnregisterGLBufferObject(cudaGraphicsResource *resource)
{
	RX_CUCHECK(cudaGraphicsUnregisterResource(resource));
}

/*!
 * foCXzXgւ̃Rs[
 * @param[in] hDst Rs[zXg(ŒsizemۂĂ邱)
 * @param[in] dSrc Rs[foCX
 * @param[in] vbo dSrcVBȌꍇCVBOIDDłȂꍇ0w
 * @param[in] size Rs[TCY(̃TCY)
 */
void CuCopyArrayFromDevice(void* hDst, const void* dSrc, cudaGraphicsResource **resource, int offset, int size)
{   
	if(resource) dSrc = CuMapGLBufferObject(resource);

	RX_CUCHECK(cudaMemcpy(hDst, (char*)dSrc+offset, size, cudaMemcpyDeviceToHost));
	
	if(resource) CuUnmapGLBufferObject(*resource);
}

/*!
 * zXgfoCXւ̃Rs[
 * @param[in] dDst Rs[foCX(ŒsizemۂĂ邱)
 * @param[in] hSrc Rs[zXg
 * @param[in] offset Rs[ItZbg
 * @param[in] size Rs[TCY(̃TCY)
 */
void CuCopyArrayToDevice(void* dDst, const void* hSrc, int offset, int size)
{
	RX_CUCHECK(cudaMemcpy((char*)dDst+offset, hSrc, size, cudaMemcpyHostToDevice));
}

/*!
 * Xbh
 */
void CuThreadSync(void)
{
	RX_CUCHECK(cudaThreadSynchronize());
}

/*!
 * foCXvpeB̕\
 */
void CuDeviceProp(void)
{
	int n;	//foCX
	RX_CUCHECK(cudaGetDeviceCount(&n));

	for(int i = 0; i < n; ++i){
		cudaDeviceProp dev;

		// foCXvpeB擾
		RX_CUCHECK(cudaGetDeviceProperties(&dev, i));

		printf("device %d\n", i);
		printf(" device name : %s\n", dev.name);
		printf(" total global memory : %d (MB)\n", dev.totalGlobalMem/1024/1024);
		printf(" shared memory / block : %d (KB)\n", dev.sharedMemPerBlock/1024);
		printf(" register / block : %d\n", dev.regsPerBlock);
		printf(" warp size : %d\n", dev.warpSize);
		printf(" max pitch : %d (B)\n", dev.memPitch);
		printf(" max threads / block : %d\n", dev.maxThreadsPerBlock);
		printf(" max size of each dim. of block : (%d, %d, %d)\n", dev.maxThreadsDim[0], dev.maxThreadsDim[1], dev.maxThreadsDim[2]);
		printf(" max size of each dim. of grid  : (%d, %d, %d)\n", dev.maxGridSize[0], dev.maxGridSize[1], dev.maxGridSize[2]);
		printf(" clock rate : %d (MHz)\n", dev.clockRate/1000);
		printf(" total constant memory : %d (KB)\n", dev.totalConstMem/1024);
		printf(" compute capability : %d.%d\n", dev.major, dev.minor);
		printf(" alignment requirement for texture : %d\n", dev.textureAlignment);
		printf(" device overlap : %s\n", (dev.deviceOverlap ? "ok" : "not"));
		printf(" num. of multiprocessors : %d\n", dev.multiProcessorCount);
		printf(" kernel execution timeout : %s\n", (dev.kernelExecTimeoutEnabled ? "on" : "off"));
		printf(" integrated : %s\n", (dev.integrated ? "on" : "off"));
		printf(" host memory mapping : %s\n", (dev.canMapHostMemory ? "on" : "off"));

		printf(" compute mode : ");
		if(dev.computeMode == cudaComputeModeDefault) printf("default mode (multiple threads can use) \n");
		else if(dev.computeMode == cudaComputeModeExclusive) printf("exclusive mode (only one thread will be able to use)\n");
		else if(dev.computeMode == cudaComputeModeProhibited) printf("prohibited mode (no threads can use)\n");
		
	}

	printf("Device with Maximum GFLOPS : %d\n", gpuGetMaxGflopsDeviceId());
}

/*!
 * thrust::exclusive_scaňĂяo
 * @param[out] dScanData scañf[^
 * @param[in] dData f[^
 * @param[in] num f[^
 */
void CuScan(unsigned int* dScanData, unsigned int* dData, unsigned int num)
{
	thrust::exclusive_scan(thrust::device_ptr<unsigned int>(dData), 
						   thrust::device_ptr<unsigned int>(dData+num),
						   thrust::device_ptr<unsigned int>(dScanData));
}

/*!
 * thrust::sort_by_keyɂnbVlɊÂ\[g
 * @param[in] dHash nbVl
 * @param[in] dIndex CfbNX(p[eBNC|SȂ)
 * @param[in] num f[^
 */
void CuSort(unsigned int *dHash, uint *dIndex, uint num)
{
	thrust::sort_by_key(thrust::device_ptr<unsigned int>(dHash),
						thrust::device_ptr<unsigned int>(dHash+num),
						thrust::device_ptr<unsigned int>(dIndex));
}




}   // extern "C"
