/*!
  @file rx_fltk_glcanvas.h
	
  @brief FLTKɂOpenGLEBhENX
 
  @author Makoto Fujisawa 
  @date   2011-09
*/

#ifndef _RX_FLTK_GLCANVAS_H_
#define _RX_FLTK_GLCANVAS_H_



//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------
#include <iostream>

// STL
#include <vector>
#include <string>

// FLTK
#include <FL/Fl.H>
#include <FL/Fl_Gl_Window.H>
#include <FL/Fl_Menu_Bar.H>
#include <FL/Fl_Spinner.H>
#include <FL/Fl_Value_Slider.H>
#include <FL/Fl_Round_Button.H>
#include <FL/Fl_Check_Button.H>

#include "rx_sph_commons.h"
#include "rx_sph_config.h"
#include "rx_fltk_widgets.h"

#include "rx_trackball.h"
#include "rx_model.h"

#include "rx_texture.h"


using namespace std;

//-----------------------------------------------------------------------------
// `/萔
//-----------------------------------------------------------------------------
class rxFlWindow;
class rxParticleSystemBase;
struct rxCell;

class rxMCMeshCPU;
class rxMCMeshGPU;

#define RX_USE_GPU


#if defined(RX_USE_GPU)
	#define RXSPH rxPBDSPH_GPU
#else
	#define RXSPH rxPBDSPH
#endif



// `tO
enum
{
	RXD_PARTICLE		= 0x0001,	//!< p[eBN
	RXD_VELOCITY		= 0x0002,	//!< x
	RXD_BBOX			= 0x0004,	//!< AABB(V~[V)
	RXD_AXIS			= 0x0008,   //!< 

	RXD_CELLS			= 0x0010,	//!< ߖTTpZ
	RXD_MESH			= 0x0020,	//!< bV
	RXD_SOLID			= 0x0040,	//!< ő
	RXD_REFRAC			= 0x0080,	//!< ܕ`

	RXD_BPARTICLE		= 0x0100,	//!< Ep[eBN
	RXD_BPARTICLE_ENV	= 0x0200,	//!< Ep[eBN`
	RXD_ZPLANE_CLIP		= 0x0400,	//!< f
	RXD_PARAMS			= 0x0800,	//!< p[^ʕ`
};
const string RX_DRAW_STR[] = {
	"Particle",					"p", 
	"Velocity",					"v", 
	"AABB (Simulation Space)", 	"b",
	"Axis",				 		"",

	"Cells", 					"d",
	"Mesh", 					"m",
	"Solid", 					"o",
	"Refrac", 					"r",
	
	"Boundary Particle",		"B",
	"Env. Boundary Prt.",		"", 
	"Cross Section",			"",
	"Params", 					"P",
	"-1"
};

// p[eBN`@
enum
{
	RXP_POINTSPRITE = 0, 
	RXP_POINT, 
	RXP_SPHERE, 
	RXP_POINT_NONE, 

	RXP_END, 
};
const string RX_PARTICLE_DRAW[] = {
	"Point Sprite",		"^1", 
	"GL_POINT", 		"^2", 
	"glutSolidSphere", 	"^3", 
	"None", 			"^4", 
	"-1"
};

// p[eBN`F@
enum
{
	RXP_COLOR_CONSTANT = 0, 
	RXP_COLOR_DENSITY, 
	RXP_COLOR_RAMP, 
	RXP_COLOR_END, 
};
const string RX_PARTICLE_COLOR[] = {
	"Constant", 	"", 
	"Density", 		"", 
	"Ramp",			"", 
	"-1"
};


// ő̕`
enum
{
	RXS_VERTEX		= 0x0001, 
	RXS_EDGE		= 0x0002, 
	RXS_FACE		= 0x0004, 
	RXS_NORMAL		= 0x0008, 
	RXS_END
};
const string RXS_STR[] = {
	"Vertex", "", 
	"Edge",   "", 
	"Face",   "", 
	"Normal", "", 
	"-1"
};

// t̕\ʕ`
enum
{
	RXL_VERTEX		= 0x0001, 
	RXL_EDGE		= 0x0002, 
	RXL_FACE		= 0x0004, 
	RXL_NORMAL		= 0x0008, 
	RXL_END
};
const string RXL_STR[] = {
	"Vertex", "", 
	"Edge",   "", 
	"Face",   "", 
	"Normal", "", 
	"-1"
};


//! `̈TCY
const string RX_CANVAS_SIZE_STR[] = {
	"1920x1080",	"",
	"1280x720",		"", 
	"1024x768",		"", 
	"800x800",		"", 
	"800x600",		"", 
	"640x480",		"", 
	"-1", 
};


//! SPHݒ
enum SettingMode
{
	RX_SPH_MESH			= 0x0001,	// bV
	RX_SPH_INPUT		= 0x0002,		// p[eBNf[^o
	RX_SPH_OUTPUT		= 0x0004,		// p[eBNf[^
	RX_SPH_MESH_OUTPUT	= 0x0008, // bVo
	RX_SPH_INLET		= 0x0010,		// E

	RX_SPH_END			= 0x0040, 
};

//! sbNꂽIuWFNgɊւ
struct rxPickInfo
{
	GLuint name;		//!< qbgIuWFNg̖O
	float min_depth;	//!< v~eBũfvXl̍ŏl
	float max_depth;	//!< v~eBũfvXl̍ől
};

static inline bool CompFuncPickInfo(rxPickInfo a, rxPickInfo b)
{
	return a.min_depth < b.min_depth;
}

//-----------------------------------------------------------------------------
//! rxFlGLWindowNX - fltkɂOpenGLEBhE
//-----------------------------------------------------------------------------
class rxFlGLWindow : public Fl_Gl_Window
{
protected:
	int m_iWinW;					//!< `EBhE̕
	int m_iWinH;					//!< `EBhE̍
	int m_iMouseButton;				//!< }EX{^̏
	int m_iKeyMod;					//!< CL[̏
	rxTrackball m_tbView;			//!< gbN{[

	double m_fBGColor[3];			//!< wiF
	bool m_bAnimation;				//!< Aj[VON/OFF
	bool m_bFullscreen;				//!< tXN[ON/OFF

	rxFlWindow *m_pParent;			//!< eNX

	vector<rxPolygons> m_vPolys;	//!< |SIuWFNg

	// FTGL
	unsigned long m_ulFontSize;		//!< tHgTCY

	//
	// q@֘Aϐ
	//
	rxParticleSystemBase *m_pPS;	//!< SPH
	double m_fDt;					//!< ^CXebv
	double m_fGravity;				//!< d͉x

	int m_iCurrentStep;				//!< ݂̃Xebv
	bool m_bPause;					//!< V~[Ṽ|[YtO

	// V[
	rxSceneConfig m_Scene;


	// p[eBNo
	string m_strSphOutputName0 ;
	string m_strSphOutputHeader;

	// p[eBN
	string m_strSphInputName0 ;
	string m_strSphInputHeader;


	//
	// bV
	//
	uint m_iNumVrts, m_iNumTris;	//!< ꂽbV̒_ƃbV
	int m_iVertexStore;				//!< TvO{[ɑ΂\z钸_(nx*ny*store)

	int m_iMeshMaxN;				//!< bVObh(EƂ̕)
	int m_iMeshN[3];				//!< bVObh

	Vec3 m_vMeshBoundaryExt;		//!< bVE{bNX̊eӂ̒1/2
	Vec3 m_vMeshBoundaryCen;		//!< bVE{bNX̒SW

	rxPolygons m_Poly;				//!< bV
	rxMaterialOBJ m_matPoly;

	GLuint m_uVrtVBO;				//!< bV_(VBO)
	GLuint m_uTriVBO;				//!< bV|S(VBO)
	GLuint m_uNrmVBO;				//!< bV_@(VBO)
	int m_iDimVBO;

	// bVo
	int m_iSaveMeshSpacing;

	// wi摜
	bool m_bUseCubeMap;				//!< L[u}bvgptO
	rxCubeMapData m_CubeMap;		//!< L[u}bv

	// bV
	rxMCMeshCPU *m_pMCMeshCPU;
	rxMCMeshGPU *m_pMCMeshGPU;


	int m_iPickedParticle;			//!< }EXsbNꂽp[eBN
	int m_iSelBufferSize;			//!< ZNVobt@̃TCY
	GLuint* m_pSelBuffer;			//!< ZNVobt@


public:
	// `tO
	int m_iDraw;					//!< `tO
	int m_iDrawPS;					//!< p[eBN`@
	int m_iColorType;				//!< p[eBN`掞̐F
	int m_iSolidDraw;				//!< ő̕`
	int m_iLiquidDraw;				//!< t̕\ʕ`

	// V~[Vݒ
	int m_iSimuSetting;		//!< ~[VݒtO
	double m_fVScale;				//!< xNg`掞̃XP[
	double m_fMeshThr;				//!< A֐bV臒l

	double m_fClipPlane[2];			//!< Nbv(zɐȕ)

	// V[Xg
	vector<string> m_vSceneTitles;	//!< V[t@CXg
	int m_iCurrentSceneIdx;			//!< ݂̃V[t@C

	// 摜o
	int m_iSaveImageSpacing;		//!< 摜ۑԊu(=-1ȂۑȂ)

public:
	//! RXgN^
	rxFlGLWindow(int x, int y, int w, int h, const char* l, void *parent);

	//! fXgN^
	~rxFlGLWindow();

public:
	// OpenGL
	void InitGL(void);

	// OpenGL`
	void Projection(void);
	vector<string> SetDrawString(void);
	void ReDisplay(void);

	// GUIR[obN
	void Display(void);
	void Resize(int w, int h);
	void Mouse(int button, int state, int x, int y);
	void Motion(int x, int y);
	void PassiveMotion(int x, int y);
	void Idle(void);
	void Timer(void);
	void Keyboard(int key, int x, int y);
	void SpecialKey(int key, int x, int y);

	// }EXsbNp
	static void Projection_s(void* x);
	static void DisplayForPick_s(void* x);
	void DisplayForPick(void);
	bool PickParticle(int x, int y);

	// _
	void InitView(void);

	// Aj[V
	static void OnTimer_s(void* x);
	static void OnIdle_s(void* x);
	
	// Aj[V؂ւ
	bool SwitchIdle(int on);

	// tXN[؂ւ
	void SwitchFullScreen(int win = 1);
	int  IsFullScreen(void);


	// t@Co
	void OpenFile(const string &fn);
	void SaveFile(const string &fn);
	void SaveDisplay(const string &fn);
	void SaveDisplay(const int &stp);
	
	void SaveMesh(const string fn, rxPolygons &polys);
	void SaveMesh(const int &stp, rxPolygons &polys);


public:
	// bV
	bool CalMeshSPH(int nmax, double thr = 1000.0);
	bool ResetMesh(void);
	void SetMeshCreation(void);
	RXREAL GetImplicitSPH(double x, double y, double z);

protected:
	bool calMeshSPH_CPU(int nmax, double thr = 1000.0);
	bool calMeshSPH_GPU(int nmax, double thr = 1000.0);
	
	// }EXsbN
	vector<rxPickInfo> selectHits(GLint nhits, GLuint buf[]);
	vector<rxPickInfo> pick(int x, int y, int w, int h);

public:
	// SPH
	void InitSPH(rxSceneConfig &sph_scene);
	void InitSPH(void){ InitSPH(m_Scene); }
	void AddSphere(void);

	void StepPS(double dt);
	void ComputeFPS(void);

	void DivideString(const string &org, vector<string> &div);

	void DrawParticleVector(RXREAL *prts, RXREAL *vels, int n, int d, double *c0, double *c1, double len = 0.1);
	void DrawParticlePoints(unsigned int vbo, int n, unsigned int color_vbo = 0, RXREAL *data = 0, int offset = 0);

	void CreateVBO(GLuint* vbo, unsigned int size);
	void DeleteVBO(GLuint* vbo);

	void DrawLiquidSurface(void);
	void SetParticleColorType(int type);

	void RenderSphScene(void);

	bool IsArtificialPressureOn(void);

private:
	//! `R[obN֐̃I[o[Ch
	void draw(void)
	{
		if(!context_valid()) InitGL();
		if(!valid()) Resize(w(), h());
		Display();    // OpenGL`
	}

	//! TCYR[obN֐̃I[o[Ch
	void resize(int x_, int y_, int w_, int h_)
	{
		Fl_Gl_Window::resize(x_, y_, w_, h_);
		//Resize(w_, h_);
	}


public:
	// Cxgnh
	int handle(int e);	// handle֐̃I[o[Ch

	// j[Cxgnh
	static void OnMenuDraw_s(Fl_Widget *widget, void* x);
	inline void OnMenuDraw(double val, string label);

	static void OnMenuSimulation_s(Fl_Widget *widget, void* x);
	inline void OnMenuSimulation(double val, string label);

	static void OnMenuParticle_s(Fl_Widget *widget, void* x);
	inline void OnMenuParticle(double val, string label);
	inline void OnMenuParticleColor(double val, string label);
	inline void OnMenuParticleDraw(double val, string label);

	static void OnMenuMesh_s(Fl_Widget *widget, void* x);
	inline void OnMenuMesh(double val, string label);
	inline void OnMenuMeshSolid(double val, string label);
	inline void OnMenuMeshLiquid(double val, string label);

	static void OnMenuScene_s(Fl_Widget *widget, void* x);
	inline void OnMenuScene(double val, string label);
};




#endif // #ifdef _RX_FLTK_GLCANVAS_H_
