/*!
  @file rx_sph.h
	
  @brief SPH@
 
  @author Makoto Fujisawa
  @date 2008-10,2011-06
*/
// FILE --rx_sph.h--

#ifndef _RX_SPH_H_
#define _RX_SPH_H_


//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------
#include "rx_sph_commons.h"

#include "rx_ps.h"			// p[eBNVXeNX
#include "rx_nnsearch.h"	// ObhɂߖTT

#include "rx_sph_solid.h"

#include "rx_kernel.h"

#include "rx_cu_common.cuh"



//-----------------------------------------------------------------------------
// `
//-----------------------------------------------------------------------------

//#define GL_REAL GL_DOUBLE
#define GL_REAL GL_FLOAT

// Ԍv
class rxTimerAvg;
extern rxTimerAvg g_Time;

#define RX_USE_TIMER

#ifdef RX_USE_TIMER
#define RXTIMER_CLEAR g_Time.ClearTime()
#define RXTIMER_RESET g_Time.ResetTime()
#define RXTIMER(x) g_Time.Split(x)
#define RXTIMER_PRINT g_Time.Print()
#define RXTIMER_STRING(x) g_Time.PrintToString(x)
#else
#define RXTIMER_CLEAR
#define RXTIMER_RESET
#define RXTIMER(x) 
#define RXTIMER_PRINT
#define RXTIMER_STRING(x)
#endif

const RXREAL KOL2 = (RXREAL)0.561231024;	// 2^(-5/6)

// O[oϐ̐錾


inline RXREAL3 MAKE_RXREAL3(RXREAL x, RXREAL y, RXREAL z)
{
	return make_float3(x, y, z);
}
inline RXREAL2 MAKE_RXREAL2(RXREAL x, RXREAL y)
{
	return make_float2(x, y);
}
inline RXREAL3 MAKE_FLOAT3V(Vec3 x)
{
	return make_float3((FLOAT)x[0], (FLOAT)x[1], (FLOAT)x[2]);
}



//! SPHV[̃p[^
struct rxEnviroment
{
	#define MAX_DELETE_REGIONS 64

	int max_particles;			//!< őp[eBN
	Vec3 boundary_cen;			//!< E̒S
	Vec3 boundary_ext;			//!< Ȇ傫(eӂ̒1/2)
	RXREAL dens;				//!< x
	RXREAL mass;				//!< p[eBN̎
	RXREAL kernel_particles;	//!< Lahȓ̃p[eBN
	RXREAL dt;					//!< ԃXebv
	RXREAL viscosity;			//!< SW
	RXREAL gas_k;				//!< KX萔

	int use_inlet;				//!< E̗L

	RXREAL epsilon;				//!< CFM̊ɘaW
	RXREAL eta;					//!< xϓe
	int min_iter;				//!< Rrŏ
	int max_iter;				//!< Rrő吔

	int use_ap;					//!< lHON/OFF (0 or 1)
	RXREAL ap_k;				//!< lH͂̂߂̌Wk ({)
	RXREAL ap_n;				//!< lH͂̂߂̌Wn (n)
	RXREAL ap_q;				//!< lH͌vZ̊J[llvZpW(Lahɑ΂W, [0,1])

	// \ʃbV
	Vec3 mesh_boundary_cen;		//!< bVE̒S
	Vec3 mesh_boundary_ext;		//!< bVȆ傫(eӂ̒1/2)
	int mesh_vertex_store;		//!< _|S\Ƃ̌W
	int mesh_max_n;				//!< MC@pObh̍ő啪

	rxEnviroment()
	{
		max_particles = 50000;
		boundary_cen = Vec3(0.0);
		boundary_ext = Vec3(2.0, 0.8, 0.8);
		dens = (RXREAL)998.29;
		mass = (RXREAL)0.04;
		kernel_particles = (RXREAL)20.0;
		dt = 0.01;
		viscosity = 1.0e-3;
		gas_k = 3.0;

		mesh_vertex_store = 10;
		use_inlet = 0;
		mesh_max_n = 128;

		epsilon = 0.001;
		eta = 0.05;
		min_iter = 2;
		max_iter = 10;

		use_ap = true;
		ap_k = 0.1;
		ap_n = 4.0;
		ap_q = 0.2;
	}
};



//! \ʃp[eBN
struct rxSurfaceParticle
{
	Vec3 pos;					//!< SW
	Vec3 nrm;					//!< @
	Vec3 vel;					//!< x
	RXREAL d;					//!< TS̋
	int idx;					//!< p[eBNCfbNX
};

extern double g_fSurfThr[2];


//-----------------------------------------------------------------------------
// MARK:rxPBDSPHNX̐錾
//  - Miles Macklin and Matthias Muller, "Position Based Fluids", Proc. SIGGRAPH 2013, 2013. 
//  - http://blog.mmacklin.com/publications/
//-----------------------------------------------------------------------------
class rxPBDSPH : public rxParticleSystemBase
{
private:
	// p[eBN
	RXREAL *m_hFrc;					//!< p[eBNɂ
	RXREAL *m_hDens;				//!< p[eBNx

	RXREAL *m_hS;					//!< Scaling factor for CFM
	RXREAL *m_hDp;					//!< ʒuC

	RXREAL *m_hPredictPos;			//!< \ʒu
	RXREAL *m_hPredictVel;			//!< \x

	RXREAL *m_hSb;					//!< Ep[eBNScaling factor

	// EEő
	rxSolid *m_pBoundary;			//!< V~[VԂ̋E
	vector<rxSolid*> m_vSolids;		//!< ő̕
	RXREAL *m_hVrts;				//!< ő̃|S̒_
	int m_iNumVrts;					//!< ő̃|S̒_
	int *m_hTris;					//!< ő̃|S
	int m_iNumTris;					//!< ő̃|S̐

	// ԕiq֘A
	rxNNGrid *m_pNNGrid;			//!< ObhɂߖTT
	vector< vector<rxNeigh> > m_vNeighs;	//!< ߖTp[eBN

	rxNNGrid *m_pNNGridB;			//!< Ep[eBNpObh
	vector< vector<rxNeigh> > m_vNeighsB;	//!< EߖTp[eBN


	// qp[^
	uint m_iKernelParticles;		//!< J[l̃p[eBN
	RXREAL m_fRestDens, m_fMass;	//!< xC
	RXREAL m_fEffectiveRadius;		//!< La
	RXREAL m_fKernelRadius;			//!< J[l̉e͈

	RXREAL m_fRestConstraint;		//!< XP[Ot@N^̕ꍀ

	// V~[Vp[^
	RXREAL m_fViscosity;			//!< SW

	RXREAL m_fEpsilon;				//!< CFM̊ɘaW
	RXREAL m_fEta;					//!< xϓ
	int m_iMinIterations;			//!< Rrŏ
	int m_iMaxIterations;			//!< Rrő唽

	bool m_bArtificialPressure;		//!< NX^Oh߂Artificial PressureǉtO
	RXREAL m_fApK;					//!< lH͂̂߂̌Wk
	RXREAL m_fApN;					//!< lH͂̂߂̌Wn
	RXREAL m_fApQ;					//!< lH͌vZ̊J[llvZpW(Lahɑ΂W, [0,1])


	// J[l֐̌vZ̍ۂɗp萔W
	double m_fAw;
	double m_fAg;
	double m_fAl;

	// J[l֐
	double (*m_fpW)(double, double, double);
	Vec3 (*m_fpGW)(double, double, double, Vec3);
	double (*m_fpLW)(double, double, double, double);

protected:
	rxPBDSPH(){}

public:
	//! RXgN^
	rxPBDSPH(bool use_opengl);

	//! fXgN^
	virtual ~rxPBDSPH();

	// p[eBNa
	float GetEffectiveRadius(){ return m_fEffectiveRadius; }

	// ߖTp[eBN
	uint* GetNeighborList(const int &i, int &n);

public:
	//
	// z֐
	//
	// p[eBNf[^
	virtual RXREAL* GetParticle(void);
	virtual RXREAL* GetParticleDevice(void);
	virtual void UnmapParticle(void){}

	// V~[VXebv
	virtual bool Update(RXREAL dt, int step = 0);

	// V[̐ݒ
	virtual void SetPolygonObstacle(const vector<Vec3> &vrts, const vector<Vec3> &nrms, const vector< vector<int> > &tris, Vec3 vel);
	virtual void SetBoxObstacle(Vec3 cen, Vec3 ext, Vec3 ang, Vec3 vel, int flg);
	virtual void SetSphereObstacle(Vec3 cen, double rad, Vec3 vel, int flg);

	// zXg<->VBOԓ]
	virtual RXREAL* GetArrayVBO(rxParticleArray type, bool d2h = true, int num = -1);
	virtual void SetArrayVBO(rxParticleArray type, const RXREAL* data, int start, int count);
	virtual void SetColorVBO(int type, int picked);

	// A֐lvZ
	virtual double GetImplicit(double x, double y, double z);
	virtual void CalImplicitField(int n[3], Vec3 minp, Vec3 d, RXREAL *hF);
	virtual void CalImplicitFieldDevice(int n[3], Vec3 minp, Vec3 d, RXREAL *dF);

	// SPHo
	virtual void OutputSetting(string fn);

	// `֐
	virtual void DrawCell(int i, int j, int k);
	virtual void DrawCells(Vec3 col, Vec3 col2, int sel = 0);
	virtual void DrawObstacles(int drw);


public:
	// SPH
	void Initialize(const rxEnviroment &env);
	void Allocate(int max_particles);
	void Finalize(void);

	// ߖT擾
	void GetNearestNeighbors(Vec3 pos, vector<rxNeigh> &neighs, RXREAL h = -1.0);
	void GetNearestNeighbors(int idx, RXREAL *p, vector<rxNeigh> &neighs, RXREAL h = -1.0);

	void GetNearestNeighborsB(Vec3 pos, vector<rxNeigh> &neighs, RXREAL h = -1.0);

	// ZɃp[eBNi[
	virtual void SetParticlesToCell(void);
	virtual void SetParticlesToCell(RXREAL *prts, int n, RXREAL h);

	// ^{[ɂA֐l
	double CalColorField(double x, double y, double z);

	// ZɃ|Si[
	virtual void SetPolygonsToCell(void);

	int  GetPolygonsInCell(int gi, int gj, int gk, set<int> &polys);
	bool IsPolygonsInCell(int gi, int gj, int gk);

	// l͍
	bool& GetArtificialPressure(void){ return m_bArtificialPressure; }


protected:
	// CPUɂSPHvZ
	void calDensity(const RXREAL *pos, RXREAL *dens, RXREAL h);
	void calForceExtAndVisc(const RXREAL *pos, const RXREAL *vel, const RXREAL *dens, RXREAL *frc, RXREAL h);

	// Scaling factořvZ
	void calScalingFactor(const RXREAL *ppos, RXREAL *pdens, RXREAL *pscl, RXREAL h, RXREAL dt);

	// Tʂ̌vZ
	void calPositionCorrection(const RXREAL *ppos, const RXREAL *pscl, RXREAL *pdp, RXREAL h, RXREAL dt);

	// rest density̌vZ
	RXREAL calRestDensity(RXREAL h);

	// Ep[eBN̑̐ςvZ
	void calBoundaryVolumes(const RXREAL *bpos, RXREAL *bvol, RXREAL mass, uint n, RXREAL h);

	// ԃXebv̏C
	RXREAL calTimeStep(RXREAL &dt, RXREAL eta_avg, const RXREAL *pfrc, const RXREAL *pvel, const RXREAL *pdens);

	// ʒuƑx̍XV
	void integrate(const RXREAL *pos, const RXREAL *vel, const RXREAL *dens, const RXREAL *acc, 
				   RXREAL *pos_new, RXREAL *vel_new, RXREAL dt);
	void integrate2(const RXREAL *pos, const RXREAL *vel, const RXREAL *dens, const RXREAL *acc, 
				    RXREAL *pos_new, RXREAL *vel_new, RXREAL dt);

	// Փ˔
	int calCollisionPolygon(uint grid_hash, Vec3 &pos0, Vec3 &pos1, Vec3 &vel, RXREAL dt);
	int calCollisionSolid(Vec3 &pos0, Vec3 &pos1, Vec3 &vel, RXREAL dt);
};




//-----------------------------------------------------------------------------
// MARK:rxPBDSPH_GPUNX̐錾
//  - Miles Macklin and Matthias Muller, "Position Based Fluids", Proc. SIGGRAPH 2013, 2013. 
//  - http://blog.mmacklin.com/publications/
//-----------------------------------------------------------------------------
class rxPBDSPH_GPU : public rxParticleSystemBase
{
private:
	//
	// oϐ(GPUϐ)
	//
	RXREAL *m_dPos;			//!< p[eBNʒu
	RXREAL *m_dVel;			//!< p[eBNx

	RXREAL *m_dFrc;			//!< p[eBNɂ
	RXREAL *m_dDens;		//!< p[eBNx

	RXREAL *m_dPosB;		//!< Ep[eBN
	RXREAL *m_dVolB;		//!< Ep[eBN̑̐

	RXREAL *m_dS;			//!< Scaling factor for CFM
	RXREAL *m_dDp;			//!< ʒuC
	RXREAL *m_dPredictPos;	//!< \ʒu
	RXREAL *m_dPredictVel;	//!< \x

	RXREAL *m_dSb;			//!< Ep[eBNScaling factor

	RXREAL *m_dErr;			//!< xϓl	
	RXREAL *m_dErrScan;		//!< xϓlScan

	cudaGraphicsResource *m_pPosResource;	//!< OpenGL(VBO)-CUDAԂ̃f[^]߂̃nh

	// \ʃbV
	RXREAL *m_dVrts;		//!< ő̃bV_
	int    *m_dTris;		//!< ő̃bV


	// V~[Vp[^
	rxSimParams m_params;	//!< V~[Vp[^(GPUւ̃f[^np)
	uint3 m_gridSize;		//!< ߖTTObh̊e̕
	uint m_numGridCells;	//!< ߖTTObh
	
	// ԕ(GPU)
	rxParticleCell m_dCellData;			//!< ߖTTObh
	uint m_gridSortBits;				//!< nbVlɂ\[g̊

	rxParticleCell m_dCellDataB;		//!< Ep[eBNpߖTTObh
	uint3 m_gridSizeB;					//!< Ep[eBNpߖTTObh̊e̕

	//
	// oϐ(CPUϐ)
	//
	RXREAL *m_hFrc;			//!< p[eBNɂ
	RXREAL *m_hDens;		//!< p[eBNx

	RXREAL *m_hS;			//!< Scaling factor for CFM
	RXREAL *m_hDp;			//!< ʒuC
	RXREAL *m_hPredictPos;	//!< \ʒu
	RXREAL *m_hPredictVel;	//!< \x

	RXREAL *m_hSb;			//!< Ep[eBNScaling factor

	// \ʃbV
	rxPolygons m_Poly;
	vector<RXREAL> m_vVrts;	//!< ő̃bV_
	int m_iNumVrts;			//!< ő̃bV_
	vector<int> m_vTris;	//!< ő̃bV
	int m_iNumTris;			//!< ő̃bV

	RXREAL m_fEpsilon;				//!< CFM̊ɘaW
	RXREAL m_fEta;					//!< xϓ
	int m_iMinIterations;			//!< Rrŏ
	int m_iMaxIterations;			//!< Rrő唽

	bool m_bArtificialPressure;		//!< NX^Oh߂Artificial PressureǉtO

	vector<rxSolid*> m_vSolids;		//!< ő̕(ő̃p[eBNp)

protected:
	rxPBDSPH_GPU(){}

public:
	//! RXgN^
	rxPBDSPH_GPU(bool use_opengl);

	//! fXgN^
	~rxPBDSPH_GPU();

	// p[eBNa
	float GetEffectiveRadius(){ return m_params.EffectiveRadius; }

	// ߖTp[eBN
	uint* GetNeighborList(const int &i, int &n);

	// V~[Vp[^
	rxSimParams GetParams(void){ return m_params; }
	void UpdateParams(void);

public:
	//
	// z֐
	//
	// p[eBNf[^
	virtual RXREAL* GetParticle(void);
	virtual RXREAL* GetParticleDevice(void);
	virtual void UnmapParticle(void);

	// V~[VXebv
	virtual bool Update(RXREAL dt, int step = 0);

	// V[̐ݒ
	virtual void SetPolygonObstacle(const vector<Vec3> &vrts, const vector<Vec3> &nrms, const vector< vector<int> > &tris, Vec3 vel);
	virtual void SetPolygonObstacle(const string &filename, Vec3 cen, Vec3 ext, Vec3 ang, Vec3 vel);
	virtual void SetBoxObstacle(Vec3 cen, Vec3 ext, Vec3 ang, Vec3 vel, int flg);
	virtual void SetSphereObstacle(Vec3 cen, double rad, Vec3 vel, int flg);

	// zXg<->VBOԓ]
	virtual RXREAL* GetArrayVBO(rxParticleArray type, bool d2h = true, int num = -1);
	virtual void SetArrayVBO(rxParticleArray type, const RXREAL* data, int start, int count);
	virtual void SetColorVBO(int type, int picked);


	// A֐lvZ
	virtual double GetImplicit(double x, double y, double z);
	virtual void CalImplicitField(int n[3], Vec3 minp, Vec3 d, RXREAL *hF);
	virtual void CalImplicitFieldDevice(int n[3], Vec3 minp, Vec3 d, RXREAL *dF);

	// SPHo
	virtual void OutputSetting(string fn);

	// `֐
	virtual void DrawCell(int i, int j, int k);
	virtual void DrawCells(Vec3 col, Vec3 col2, int sel = 0);
	virtual void DrawObstacles(int drw);

protected:
	void setObjectToCell(RXREAL *p, RXREAL *v);


public:
	// SPH
	void Initialize(const rxEnviroment &env);
	void Allocate(int max_particles);
	void Finalize(void);

	// ZɃp[eBNi[
	virtual void SetParticlesToCell(void);
	virtual void SetParticlesToCell(RXREAL *prts, int n, RXREAL h);
	virtual void SetPolygonsToCell(void);

	int  GetPolygonsInCell(int gi, int gj, int gk, vector<int> &polys){ return 0; }
	bool IsPolygonsInCell(int gi, int gj, int gk){ return false; }

	void CalMaxDensity(int k);

	// l͍
	bool& GetArtificialPressure(void){ return m_bArtificialPressure; }

	// Ep[eBN̏
	virtual void InitBoundary(void);


protected:
	// rest density̌vZ
	RXREAL calRestDensity(RXREAL h);

	// Z̏ݒ
	void setupCells(rxParticleCell &cell, uint3 &gridsize, double &cell_width, Vec3 vMin, Vec3 vMax, double h);

	// ObhnbV̌vZ
	uint calGridHash(uint x, uint y, uint z);
	uint calGridHash(Vec3 pos);

	// |S𕪊ZɊi[
	void setPolysToCell(RXREAL *vrts, int nv, int* tris, int nt);

	// |S_̒
	bool fitVertices(const Vec3 &ctr, const Vec3 &sl, vector<Vec3> &vec_set);
};



#endif	// _SPH_H_

