/*!
  @file rx_sph_solid.h
	
  @brief SPHpő̒`
 
  @author Makoto Fujisawa
  @date 2008-12
*/

#ifndef _RX_SPH_SOLID_H_
#define _RX_SPH_SOLID_H_


//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------


#include "rx_sph_commons.h"
#include "rx_model.h"
#include "rx_nnsearch.h"	// ObhɂߖTT
#include "rx_mesh.h"

// SPHJ[l
#include "rx_kernel.h"


//-----------------------------------------------------------------------------
// `E萔
//-----------------------------------------------------------------------------

const int RX_DISPMAP_N = 128;


//-----------------------------------------------------------------------------
// MARK:rxCollisionInfoNX
//-----------------------------------------------------------------------------
class rxCollisionInfo
{
protected:
	Vec3 m_vContactPoint;	//!< Փ˓_
	Vec3 m_vNormal;			//!< (Փ˓_ł)@
	double m_fDepth;		//!< ߂荞ݗ

	Vec3 m_vVelocity;		//!< Փ˓_̑x

public:
	//! ftHgRXgN^
	rxCollisionInfo()
	  : m_vContactPoint(Vec3(0.0)), 
		m_vNormal(Vec3(0.0)), 
		m_fDepth(0.0), 
		m_vVelocity(Vec3(0.0))
	{
	}

	//! RXgN^
	rxCollisionInfo(const Vec3 &contact_point, 
					const Vec3 &normal = Vec3(0.0), 
					const double &depth = 0.0, 
					const Vec3 &veloc = Vec3(0.0))
	  : m_vContactPoint(contact_point), 
		m_vNormal(normal), 
		m_fDepth(depth), 
		m_vVelocity(veloc)
	{
	}

	//! fXgN^
	~rxCollisionInfo(){}

public:
	const Vec3& Contact() const { return m_vContactPoint; }
	Vec3& Contact(){ return m_vContactPoint; }

	const Vec3& Normal() const { return m_vNormal; }
	Vec3& Normal(){ return m_vNormal; }

	const double& Penetration() const { return m_fDepth; }
	double& Penetration(){ return m_fDepth; }

	const Vec3& Velocity() const { return m_vVelocity; }
	Vec3& Velocity(){ return m_vVelocity; }
};




//-----------------------------------------------------------------------------
// MARK:rxSolid : ő̃IuWFNgNX
//-----------------------------------------------------------------------------
class rxSolid
{
public:
	enum{
		RXS_SPHERE, 
		RXS_AABB, 
		RXS_OPEN_BOX, 
		RXS_POLYGON, 
		RXS_IMPLICIT, 
		RXS_OTHER = -1, 
	};

protected:
	Vec3 m_vMassCenter;	//!< dSW
	Vec3 m_vVelocity;	//!< ő̑x
	rxMatrix4 m_matRot;	//!< p
	rxMatrix4 m_matRotInv;	//!< p

	int m_iName;

	RXREAL m_fOffset;

	bool m_bFix;		//!< ŒtO
	bool m_bSP;			//!< ő̃p[eBNtO

	int m_iSgn;			//!< :-1, IuWFNg:1

public:
	rxSolid() : m_iName(RXS_OTHER)
	{
		m_bFix = true;
		m_vMassCenter = Vec3(0.0);
		m_vVelocity = Vec3(0.0);
		m_iSgn = 1;
		m_fOffset = (RXREAL)(0.0);
		m_bSP = true;

		m_matRot.MakeIdentity();
	}

	//
	// z֐
	//
	virtual bool GetDistance(const Vec3 &pos, rxCollisionInfo &col) = 0;	//!< ֐vZ
	virtual bool GetDistanceR(const Vec3 &pos, const double &r, rxCollisionInfo &col) = 0;

	virtual bool GetDistance(const Vec3 &pos0, const Vec3 &pos1, rxCollisionInfo &col) = 0;	//!< ֐vZ
	virtual bool GetDistanceR(const Vec3 &pos0, const Vec3 &pos1, const double &r, rxCollisionInfo &col) = 0;

	virtual bool GetCurvature(const Vec3 &pos, double &k) = 0;	//!< ֐̋ȗvZ

	virtual void Draw(int drw) = 0;				//!< OpenGLł̕`
	virtual void SetGLMatrix(void) = 0;							//!< OpenGLϊs̓Kp

	virtual Vec3 GetMin(void) = 0;
	virtual Vec3 GetMax(void) = 0;


	//
	// \ʏɃp[eBNzu
	//
	static Vec4 GetImplicitG_s(void* ptr, double x, double y, double z);
	inline Vec4 GetImplicitG(Vec3 pos);
	static RXREAL GetImplicit_s(void* ptr, double x, double y, double z);
	inline RXREAL GetImplicit(Vec3 pos);

	int GenerateParticlesOnSurf(RXREAL rad, RXREAL **ppos);

	static bool GetDistance_s(Vec3 pos, rxCollisionInfo &col, void* x);

	//
	// 擾Eݒ֐
	//
	inline Vec3 CalLocalCoord(const Vec3 &pos);			//!< O[oő̃[Jւ̍Wϊ
	inline Vec3 CalGlobalCoord(const Vec3 &pos);		//!< ő̃[JO[oւ̍Wϊ

	inline Vec3 GetPosition(void);						//!< ő̏dSʒu̎擾
	inline void SetPosition(const Vec3 &pos);			//!< ő̏dSʒu̐ݒ

	inline rxMatrix4 GetMatrix(void);					//!< ]s̎擾
	inline void SetMatrix(const rxMatrix4 &mat);		//!< ]s̐ݒ
	inline void SetMatrix(double mat[16]);				//!< ]s̐ݒ

	inline Vec3 GetVelocityAtGrobal(const Vec3 &pos);	//!< ̍Wľő̑x̎擾
	inline void SetVelocity(const Vec3 &vec);			//!< 

	inline bool GetFix(void) const { return m_bFix; }		//!< ŒtO̎擾
	inline void SetFix(bool fix){ m_bFix = fix; }			//!< ŒtO̐ݒ

	inline bool& IsSolidParticles(void){ return m_bSP; }	//!< ő̃p[eBNtO̎擾

	inline bool RigidSimulation(const double &dt);		//!< ̃V~[V

	inline const int& Name() const { return m_iName; }
	inline int& Name(){ return m_iName; }
};



//-----------------------------------------------------------------------------
// MARK:rxSolidBox : 
//-----------------------------------------------------------------------------
class rxSolidBox : public rxSolid
{
protected:
	Vec3 m_vMax, m_vMin;	//!< őWCŏW(S̑Βl)
	Vec3 m_vC[2];

public:
	// RXgN^
	rxSolidBox(Vec3 minp, Vec3 maxp, int sgn)
	{
		Vec3 sl  = 0.5*(maxp-minp);
		Vec3 ctr = 0.5*(maxp+minp);

		m_vMin = -sl;
		m_vMax =  sl;
		m_vMassCenter = ctr;

		m_vC[0] = m_vMin;
		m_vC[1] = m_vMax;

		m_iSgn = sgn;

		m_iName = RXS_AABB;
	}

	virtual Vec3 GetMin(void){ return m_vMassCenter+m_vMin; }
	virtual Vec3 GetMax(void){ return m_vMassCenter+m_vMax; }

	virtual bool GetDistance(const Vec3 &pos, rxCollisionInfo &col);
	virtual bool GetDistanceR(const Vec3 &pos, const double &r, rxCollisionInfo &col);

	virtual bool GetDistance(const Vec3 &pos0, const Vec3 &pos1, rxCollisionInfo &col){ return GetDistance(pos1, col); }
	virtual bool GetDistanceR(const Vec3 &pos0, const Vec3 &pos1, const double &r, rxCollisionInfo &col){ return GetDistanceR(pos1, r, col); }

	virtual bool GetCurvature(const Vec3 &pos, double &k);
	virtual void Draw(int drw = 4);
	virtual void SetGLMatrix(void);

protected:
	bool aabb_point_dist(const Vec3 &spos, const double &h, const int &sgn, const Vec3 &vMin, const Vec3 &vMax, 
						 double &d, Vec3 &n, int &np, int &plist, double pdist[6]);

};

//-----------------------------------------------------------------------------
// rxSolidOpenBox : (J)
//-----------------------------------------------------------------------------
class rxSolidOpenBox : public rxSolid
{
protected:
	Vec3 m_vSLenIn, m_vSLenOut;

public:
	// RXgN^
	rxSolidOpenBox(Vec3 ctr, Vec3 sl_in, Vec3 sl_out, int sgn)
	{
		m_vSLenIn  = sl_in;
		m_vSLenOut = sl_out;
		m_vMassCenter = ctr;

		RXCOUT << "SLenIn  " << m_vSLenIn << endl;
		RXCOUT << "SLenOut " << m_vSLenOut << endl;

		m_iSgn = sgn;

		m_iName = RXS_OPEN_BOX;
	}

	Vec3 GetInMin(void) const { return -m_vSLenIn; }
	Vec3 GetInMax(void) const { return  m_vSLenIn; }
	Vec3 GetOutMin(void) const { return -m_vSLenOut; }
	Vec3 GetOutMax(void) const { return  m_vSLenOut; }
	Vec3 GetInSideLength(void) const { return m_vSLenIn; }
	Vec3 GetOutSideLength(void) const { return m_vSLenOut; }

	virtual Vec3 GetMin(void){ return -m_vSLenOut; }
	virtual Vec3 GetMax(void){ return  m_vSLenOut; }

	virtual bool GetDistance(const Vec3 &pos, rxCollisionInfo &col);
	virtual bool GetDistanceR(const Vec3 &pos, const double &r, rxCollisionInfo &col);

	virtual bool GetDistance(const Vec3 &pos0, const Vec3 &pos1, rxCollisionInfo &col){ return GetDistance(pos1, col); }
	virtual bool GetDistanceR(const Vec3 &pos0, const Vec3 &pos1, const double &r, rxCollisionInfo &col){ return GetDistanceR(pos1, r, col); }

	virtual bool GetCurvature(const Vec3 &pos, double &k);
	virtual void Draw(int drw = 4);
	virtual void SetGLMatrix(void);
};



//-----------------------------------------------------------------------------
// rxSolidSphere : 
//-----------------------------------------------------------------------------
class rxSolidSphere : public rxSolid
{
protected:
	double m_fRadius;		//!< a
	double m_fRadiusSqr;	//!< a̎

public:
	// RXgN^
	rxSolidSphere(Vec3 ctr, double rad, int sgn)
		: m_fRadius(rad)
	{
		m_iSgn = sgn;
		m_vMassCenter = ctr;
		m_fRadiusSqr = rad*rad;

		m_iName = RXS_SPHERE;
	}

	Vec3 GetCenter(void) const { return m_vMassCenter; }
	double GetRadius(void) const { return m_fRadius; }

	virtual Vec3 GetMin(void){ return m_vMassCenter-Vec3(m_fRadius); }
	virtual Vec3 GetMax(void){ return m_vMassCenter+Vec3(m_fRadius); }

	virtual bool GetDistance(const Vec3 &pos, rxCollisionInfo &col);
	virtual bool GetDistanceR(const Vec3 &pos, const double &r, rxCollisionInfo &col);

	virtual bool GetDistance(const Vec3 &pos0, const Vec3 &pos1, rxCollisionInfo &col){ return GetDistance(pos1, col); }
	virtual bool GetDistanceR(const Vec3 &pos0, const Vec3 &pos1, const double &r, rxCollisionInfo &col){ return GetDistanceR(pos1, r, col); }

	virtual bool GetCurvature(const Vec3 &pos, double &k){ return false; }
	virtual void Draw(int drw = 4);
	virtual void SetGLMatrix(void);
};





//-----------------------------------------------------------------------------
// rxSolidPolygon : |S
//-----------------------------------------------------------------------------
class rxSolidPolygon : public rxSolid
{
protected:
	Vec3 m_vMax, m_vMin;	//!< őWCŏW(S̑Βl)

	string m_strFilename;

	rxPolygons m_Poly;
	RXREAL *m_hVrts;				//!< ő̃|S̒_
	int m_iNumVrts;					//!< ő̃|S̒_
	int *m_hTris;					//!< ő̃|S
	int m_iNumTris;					//!< ő̃|S̐

	double m_fMaxRad;
	rxNNGrid *m_pNNGrid;			//!< ObhɂߖTT

public:
	// RXgN^ƃfXgN^
	rxSolidPolygon(const string &fn, Vec3 cen, Vec3 ext, Vec3 ang, double h, int aspect = 1);
	~rxSolidPolygon();

	virtual Vec3 GetMin(void){ return m_vMassCenter+m_vMin; }
	virtual Vec3 GetMax(void){ return m_vMassCenter+m_vMax; }

	virtual bool GetDistance(const Vec3 &pos, rxCollisionInfo &col);
	virtual bool GetDistanceR(const Vec3 &pos, const double &r, rxCollisionInfo &col);

	virtual bool GetDistance(const Vec3 &pos0, const Vec3 &pos1, rxCollisionInfo &col);
	virtual bool GetDistanceR(const Vec3 &pos0, const Vec3 &pos1, const double &r, rxCollisionInfo &col);

	virtual bool GetCurvature(const Vec3 &pos, double &k);
	virtual void Draw(int drw);
	virtual void SetGLMatrix(void);

	// Z
	int  GetPolygonsInCell(int gi, int gj, int gk, set<int> &polys);
	bool IsPolygonsInCell(int gi, int gj, int gk);

protected:
	int getDistanceToPolygon(Vec3 x, int p, double &dist);

	int readFile(const string filename, rxPolygons &polys);
	void readOBJ(const string filename, rxPolygons &polys);

	bool fitVertices(const Vec3 &ctr, const Vec3 &sl, vector<Vec3> &vec_set, int aspect = 1);

	int intersectSegmentTriangle(Vec3 P0, Vec3 P1, Vec3 V0, Vec3 V1, Vec3 V2, Vec3 &I, Vec3 &n);

};




//-----------------------------------------------------------------------------
// MARK:rxSolidNX̎
//-----------------------------------------------------------------------------
/*!
 * O[oő̃[Jւ̍Wϊ
 * @param[in] x,y,z O[oWnł̈ʒu
 * @return ő̃[JWnł̈ʒu
 */
inline Vec3 rxSolid::CalLocalCoord(const Vec3 &pos)
{
	// ̍Wő̍Wւƕϊ
	Vec3 rpos;
	rpos = pos-m_vMassCenter;
	//m_matRot.multMatrixVec(rpos);
	rpos[0] = rpos[0]*m_matRot(0,0)+rpos[1]*m_matRot(1,0)+rpos[2]*m_matRot(2,0);
	rpos[1] = rpos[0]*m_matRot(0,1)+rpos[1]*m_matRot(1,1)+rpos[2]*m_matRot(2,1);
	rpos[2] = rpos[0]*m_matRot(0,2)+rpos[1]*m_matRot(1,2)+rpos[2]*m_matRot(2,2);
	return rpos;
}
/*!
 * ő̃[JO[oւ̍Wϊ
 * @param[in] x,y,z ő̃[JWnł̈ʒu
 * @return O[oWnł̈ʒu
 */
inline Vec3 rxSolid::CalGlobalCoord(const Vec3 &pos)
{
	// ő̍W痬̍Wւƕϊ
	Vec3 fpos = pos;
	//m_matRotInv.multMatrixVec(pos, fpos);
	fpos[0] = pos[0]*m_matRotInv(0,0)+pos[1]*m_matRotInv(1,0)+pos[2]*m_matRotInv(2,0);
	fpos[1] = pos[0]*m_matRotInv(0,1)+pos[1]*m_matRotInv(1,1)+pos[2]*m_matRotInv(2,1);
	fpos[2] = pos[0]*m_matRotInv(0,2)+pos[1]*m_matRotInv(1,2)+pos[2]*m_matRotInv(2,2);
	fpos = fpos+m_vMassCenter;
	return fpos;
}

/*!
 * ő̏dSʒu̎擾
 * @return ő̏dSW(̍Wn)
 */
inline Vec3 rxSolid::GetPosition(void)
{
	return m_vMassCenter;
}

/*!
 * ő̏dSʒu̐ݒ
 * @param[in] pos ő̏dSW(̍Wn)
 */
void rxSolid::SetPosition(const Vec3 &pos)
{
	m_vMassCenter = pos;
}

/*!
 * ő̏dSʒu̎擾
 * @return ő̏dSW(̍Wn)
 */
inline rxMatrix4 rxSolid::GetMatrix(void)
{
	return m_matRot;
}


inline rxMatrix4 CalInverseMatrix(const rxMatrix4 &mat)
{
	real d = mat(0, 0)*mat(1, 1)*mat(2, 2)-mat(0, 0)*mat(2, 1)*mat(1, 2)+ 
			 mat(1, 0)*mat(2, 1)*mat(0, 2)-mat(1, 0)*mat(0, 1)*mat(2, 2)+ 
			 mat(2, 0)*mat(0, 1)*mat(1, 2)-mat(2, 0)*mat(1, 1)*mat(0, 2);

	if(d == 0) d = 1;

	return rxMatrix4( (mat(1, 1)*mat(2, 2)-mat(1, 2)*mat(2, 1))/d,
					 -(mat(0, 1)*mat(2, 2)-mat(0, 2)*mat(2, 1))/d,
					  (mat(0, 1)*mat(1, 2)-mat(0, 2)*mat(1, 1))/d,
					  0.0, 
					 -(mat(1, 0)*mat(2, 2)-mat(1, 2)*mat(2, 0))/d,
					  (mat(0, 0)*mat(2, 2)-mat(0, 2)*mat(2, 0))/d,
					 -(mat(0, 0)*mat(1, 2)-mat(0, 2)*mat(1, 0))/d,
					  0.0, 
					  (mat(1, 0)*mat(2, 1)-mat(1, 1)*mat(2, 0))/d,
					 -(mat(0, 0)*mat(2, 1)-mat(0, 1)*mat(2, 0))/d,
					  (mat(0, 0)*mat(1, 1)-mat(0, 1)*mat(1, 0))/d,
					  0.0, 
					  0.0, 0.0, 0.0, 1.0);
}


/*!
 * ő̏dSʒu̐ݒ
 * @param[in] pos ő̏dSW(̍Wn)
 */
void rxSolid::SetMatrix(const rxMatrix4 &mat)
{
	//m_matRot = mat;
	for(int i = 0; i < 3; ++i){
		for(int j = 0; j < 3; ++j){
			m_matRot(i, j) = mat(i, j);
		}
	}

	//m_matRotInv = m_matRot.Inverse();
	m_matRotInv = CalInverseMatrix(m_matRot);
}

/*!
 * ő̏dSʒu̐ݒ
 * @param[in] pos ő̏dSW(̍Wn)
 */
void rxSolid::SetMatrix(double mat[16])
{
	//m_matRot = mat;
	for(int i = 0; i < 4; ++i){
		for(int j = 0; j < 4; ++j){
			m_matRot(i, j) = mat[i+j*4];
		}
	}

	//m_matRotInv = m_matRot.Inverse();
	m_matRotInv = CalInverseMatrix(m_matRot);
}


/*!
 * O[oWlł̌ő̑x̎擾
 * @param[in] pos O[oWl
 * @return ő̑x
 */
inline Vec3 rxSolid::GetVelocityAtGrobal(const Vec3 &pos)
{
	return m_vVelocity;
}

/*!
 * ő̑xZbg
 * @param[in] vec dSx
 */
inline void rxSolid::SetVelocity(const Vec3 &vec)
{
	m_vVelocity = vec;
}


/*!
 * ̃V~[V(fix=true̎)
 * @param[in] dt ^CXebv
 */
inline bool rxSolid::RigidSimulation(const double &dt)
{
	m_vMassCenter += dt*m_vVelocity;
	return true;
}


	
	
//-----------------------------------------------------------------------------
// MARK:̑֐
//-----------------------------------------------------------------------------
inline bool GetImplicitPlane(const Vec3 &pos, double &d, Vec3 &n, Vec3 &v, const Vec3 &pn, const Vec3 &pq)
{
	d = dot(pq-pos, pn);
	n = pn;
	v = Vec3(0.0);

	return true;
}


/*!
 * ֐ȗvZ
 * @param[in] pos vZ_
 * @param[out] k ȗ
 * @param[in] fpDist ֐
 */
inline bool CalCurvature(const Vec3 &pos, double &k, bool (fpDist)(Vec3, rxCollisionInfo&, void*), void* fp = 0)
{
	k = 0.0;

	double h = 0.005;
	double x0, y0, z0;
	double p[3][3][3];
	rxCollisionInfo col;

	x0 = pos[0]-0.5*h;
	y0 = pos[1]-0.5*h;
	z0 = pos[2]-0.5*h;

//	fpDist(Vec3(x0-h, y0-h, z0-h), col, fp); p[0][0][0] = col.Penetration();
	fpDist(Vec3(x0-h, y0-h, z0  ), col, fp); p[0][0][1] = col.Penetration();
//	fpDist(Vec3(x0-h, y0-h, z0+h), col, fp); p[0][0][2] = col.Penetration();
	fpDist(Vec3(x0-h, y0  , z0-h), col, fp); p[0][1][0] = col.Penetration();
	fpDist(Vec3(x0-h, y0  , z0  ), col, fp); p[0][1][1] = col.Penetration();
	fpDist(Vec3(x0-h, y0  , z0+h), col, fp); p[0][1][2] = col.Penetration();
//	fpDist(Vec3(x0-h, y0+h, z0-h), col, fp); p[0][2][0] = col.Penetration();
	fpDist(Vec3(x0-h, y0+h, z0  ), col, fp); p[0][2][1] = col.Penetration();
//	fpDist(Vec3(x0-h, y0+h, z0+h), col, fp); p[0][2][2] = col.Penetration();

	fpDist(Vec3(x0  , y0-h, z0-h), col, fp); p[1][0][0] = col.Penetration();
	fpDist(Vec3(x0  , y0-h, z0  ), col, fp); p[1][0][1] = col.Penetration();
	fpDist(Vec3(x0  , y0-h, z0+h), col, fp); p[1][0][2] = col.Penetration();
	fpDist(Vec3(x0  , y0  , z0-h), col, fp); p[1][1][0] = col.Penetration();
	fpDist(Vec3(x0  , y0  , z0  ), col, fp); p[1][1][1] = col.Penetration();
	fpDist(Vec3(x0  , y0  , z0+h), col, fp); p[1][1][2] = col.Penetration();
	fpDist(Vec3(x0  , y0+h, z0-h), col, fp); p[1][2][0] = col.Penetration();
	fpDist(Vec3(x0  , y0+h, z0  ), col, fp); p[1][2][1] = col.Penetration();
	fpDist(Vec3(x0  , y0+h, z0+h), col, fp); p[1][2][2] = col.Penetration();

//	fpDist(Vec3(x0+h, y0-h, z0-h), col, fp); p[2][0][0] = col.Penetration();
	fpDist(Vec3(x0+h, y0-h, z0  ), col, fp); p[2][0][1] = col.Penetration();
//	fpDist(Vec3(x0+h, y0-h, z0+h), col, fp); p[2][0][2] = col.Penetration();
	fpDist(Vec3(x0+h, y0  , z0-h), col, fp); p[2][1][0] = col.Penetration();
	fpDist(Vec3(x0+h, y0  , z0  ), col, fp); p[2][1][1] = col.Penetration();
	fpDist(Vec3(x0+h, y0  , z0+h), col, fp); p[2][1][2] = col.Penetration();
//	fpDist(Vec3(x0+h, y0+h, z0-h), col, fp); p[2][2][0] = col.Penetration();
	fpDist(Vec3(x0+h, y0+h, z0  ), col, fp); p[2][2][1] = col.Penetration();
//	fpDist(Vec3(x0+h, y0+h, z0+h), col, fp); p[2][2][2] = col.Penetration();

	double px, py, pz, pxx, pyy, pzz, pxy, pyz, pxz, np;
	px = (p[2][1][1]-p[0][1][1])/(2.0*h);
	py = (p[1][2][1]-p[1][0][1])/(2.0*h);
	pz = (p[1][1][2]-p[1][1][0])/(2.0*h);

	pxx = (p[2][1][1]-2.0*p[1][1][1]+p[0][1][1])/(h*h);
	pyy = (p[1][2][1]-2.0*p[1][1][1]+p[1][0][1])/(h*h);
	pzz = (p[1][1][2]-2.0*p[1][1][1]+p[1][1][0])/(h*h);

	pxy = (p[0][0][1]+p[2][2][1]-p[0][2][1]-p[2][0][1])/(4.0*h*h);
	pxz = (p[0][1][0]+p[2][1][2]-p[0][1][2]-p[2][1][0])/(4.0*h*h);
	pyz = (p[1][0][0]+p[1][2][2]-p[1][0][2]-p[1][2][0])/(4.0*h*h);

	np = px*px+py*py+pz*pz;
	if(np > RX_FEQ_EPS){
		np = sqrt(np);

		// ȗ̌vZ
		k = (px*px*pyy-2.0*px*py*pxy+py*py*pxx+px*px*pzz-2.0*px*pz*pxz+pz*pz*pxx+py*py*pzz-2.0*py*pz*pyz+pz*pz*pyy)/(np*np*np);
	}

	k = -k;

	return true;
}

/*!
 * Poly6J[l̒lvZ
 * @param[in] r J[lS܂ł̋
 * @param[in] h La
 * @param[in] ptr ֐Ăяo|C^
 * @return J[l֐l
 */
static inline double CalKernelPoly6(double r, double h, void *ptr = 0)
{
	static double a = 0.0;
	if(a == 0.0) a = KernelCoefPoly6(h, 3, 1);	
	return KernelPoly6(r, h, a);
}

/*!
 * Poly6J[l̒lvZ(ől1ɂȂ悤ɐK)
 * @param[in] r J[lS܂ł̋
 * @param[in] h La
 * @param[in] ptr ֐Ăяo|C^
 * @return J[l֐l
 */
static inline double CalKernelPoly6r(double r, double h, void *ptr = 0)
{
	static double a = 0.0;
	static double b = 0.0;
	if(a == 0.0) a = KernelCoefPoly6(h, 3, 1);	
	if(b == 0.0) b = KernelPoly6(0.0, h, a);
	return KernelPoly6(r, h, a)/b;
}


/*!
 * (C,)Ƌ̌
 * @param[in] p,d Č_ƕ
 * @param[in] c,r ̒SƔa
 * @param[out] t1,t2 p_܂ł̋
 * @return _
 */
inline int ray_sphere(const Vec3 &p, const Vec3 &d, const Vec3 &sc, const double r, double &t1, double &t2)
{
	Vec3 q = p-sc;	// SWnł̌_W

	double a = norm2(d);
	double b = 2*dot(q, d);
	double c = norm2(q)-r*r;

	// ʎ
	double D = b*b-4*a*c;

	if(D < 0.0){ // Ȃ
		return 0;
	}
	else if(D < RX_FEQ_EPS){ // _1
		t1 = -b/(2*a);
		t2 = -1;
		return 1;
	}
	else{ // _2
		double sqrtD = sqrt(D);
		t1 = (-b-sqrtD)/(2*a);
		t2 = (-b+sqrtD)/(2*a);
		return 2;
	}

}

/*!
 * Op`Ƌ̌
 * @param[in] v0,v1,v2	Op`̒_
 * @param[in] n			Op`̖@
 * @param[in] p			ŋߖT_
 * @return 
 */
inline bool triangle_sphere(const Vec3 &v0, const Vec3 &v1, const Vec3 &v2, const Vec3 &n, 
							const Vec3 &c, const double &r, double &dist, Vec3 &ipoint)
{
	// |S܂ޕʂƋS̋
	double d = dot(v0, n);
	double l = dot(n, c)-d;

	dist = l;
	if(l > r) return false;

	// ʂƂ̍ŋߖT_W
	Vec3 p = c-l*n;

	// ߖT_Op`ǂ̔
	Vec3 n1 = cross((v0-c), (v1-c));
	Vec3 n2 = cross((v1-c), (v2-c));
	Vec3 n3 = cross((v2-c), (v0-c));

	ipoint = p;
	dist = l;
	if(dot(n1, n2) > 0 && dot(n2, n3) > 0){		// Op`
		return true;
	}
	else{		// Op`O
		// Op`̊eGbWƋ̏Փ˔
		for(int e = 0; e < 3; ++e){
			Vec3 va0 = (e == 0 ? v0 : (e == 1 ? v1 : v2));
			Vec3 va1 = (e == 0 ? v1 : (e == 1 ? v2 : v0));

			double t1, t2;
			int n = ray_sphere(va0, Unit(va1-va0), c, r, t1, t2);

			if(n){
				double le2 = norm2(va1-va0);
				if((t1 >= 0.0 && t1*t1 < le2) || (t2 >= 0.0 && t2*t2 < le2)){
					return true;
				}
			}
		}
		return false;
	}
}

/*!
 * (܂ޒ)Ɠ_̋
 * @param[in] v0,v1 ̗[_W
 * @param[in] p _̍W
 * @return 
 */
inline double segment_point_dist(const Vec3 &v0, const Vec3 &v1, const Vec3 &p)
{
	Vec3 v = Unit(v1-v0);
	Vec3 vp = p-v0;
	Vec3 vh = dot(vp, v)*v;
	return norm(vp-vh);
}

/*!
 * Ƌ̌
 * @param[in] s0,s1	̒[_
 * @param[in] sc,r   ̒SWƔa
 * @param[out] d2 Ƃ̋̓
 * @return true
 */
inline bool segment_sphere(const Vec3 &s0, const Vec3 &s1, const Vec3 &sc, const double &r, double &d2)
{
	Vec3 v = s1-s0;
	Vec3 c = sc-s0;

	double vc = dot(v, c);
	if(vc < 0){		// ̒S̎n_s0̊Oɂ
		d2 = norm2(c);
		return (d2 < r*r);	// SƎn_s0̋Ō
	}
	else{
		double v2 = norm2(v);
		if(vc > v2){	// ̒S̏I_s1̊Oɂ
			d2 = norm2(s1-sc);
			return (d2 < r*r);	// SƏI_s1̋Ō
		}
		else{			// s0s1̊Ԃɂ
			d2 = norm2((vc*v)/norm2(v)-c);
			return (d2 < r*r);	// ƋS̋Ō
		}
	}

	return false;
}





#endif	// _RX_SPH_SOLID_H_
